@@ -5,10 +5,20 @@ import time
 import warnings
 from email import charset
 
+import __main__
 from lxml import etree, objectify
 
-__version__ = "6.0.53"
+try:
+    from requests import utils as requests_utils
+except ImportError:
+    requests_utils = None
 
+__version__ = "7.0.19"
+
+os.environ.setdefault(
+    'TRYTOND_APPNAME',
+    os.path.basename(getattr(__main__, '__file__', 'trytond')))
+os.environ.setdefault('TRYTOND_TZ', os.environ.get('TZ', 'UTC'))
 os.environ['TZ'] = 'UTC'
 if hasattr(time, 'tzset'):
     time.tzset()
@@ -22,3 +32,11 @@ charset.add_charset('utf-8', charset.QP, charset.QP)
 # prevent XML vulnerabilities by default
 etree.set_default_parser(etree.XMLParser(resolve_entities=False))
 objectify.set_default_parser(objectify.makeparser(resolve_entities=False))
+
+
+def default_user_agent(name="Tryton"):
+    return f"{name}/{__version__}"
+
+
+if requests_utils:
+    requests_utils.default_user_agent = default_user_agent
