@@ -2,11 +2,11 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 
-import unittest
 import datetime
 import sys
-
+import unittest
 from decimal import Decimal
+
 from trytond import pyson
 
 
@@ -317,6 +317,21 @@ class PYSONTestCase(unittest.TestCase):
                 pyson.DateTime(2020, 1, 1, 0, 0, 0, 0), 90000))
         self.assertTrue(pyson.PYSONDecoder().decode(eval))
 
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.TimeDelta(5, 2),
+                pyson.TimeDelta(4, 2)))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.TimeDelta(4, 2),
+                pyson.TimeDelta(5, 2)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.TimeDelta(5, 2),
+                datetime.timedelta(4, 2)))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
         eval = pyson.PYSONEncoder().encode(
             pyson.Greater(pyson.Eval('i', 0), 0))
         self.assertTrue(pyson.PYSONDecoder({'i': 1}).decode(eval))
@@ -409,12 +424,11 @@ class PYSONTestCase(unittest.TestCase):
             'e': 'bar',
             })
 
-        if not sys.flags.optimize:
-            self.assertRaises(AssertionError, pyson.If, True, 'foo', False)
-
         self.assertEqual(pyson.If(True, 'foo', 'bar').types(),
             set([str]))
         self.assertEqual(pyson.If(True, False, True).types(), set([bool]))
+        self.assertEqual(
+            pyson.If(True, 'foo', False).types(), set([bool, str]))
 
         eval = pyson.PYSONEncoder().encode(pyson.If(True, 'foo', 'bar'))
         self.assertEqual(pyson.PYSONDecoder().decode(eval), 'foo')
@@ -475,12 +489,20 @@ class PYSONTestCase(unittest.TestCase):
         eval = pyson.PYSONEncoder().encode(pyson.In('foo', {'foo': 'bar'}))
         self.assertTrue(pyson.PYSONDecoder().decode(eval))
 
+        eval = pyson.PYSONEncoder().encode(pyson.In(
+            pyson.Eval('test', 'foo'), {'foo': 'bar'}))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
         eval = pyson.PYSONEncoder().encode(pyson.In('1', {1: 'bar'}))
         self.assertTrue(pyson.PYSONDecoder().decode(eval))
 
         eval = pyson.PYSONEncoder().encode(pyson.In('test', {'foo': 'bar'}))
         self.assertFalse(pyson.PYSONDecoder().decode(eval))
 
+        eval = pyson.PYSONEncoder().encode(pyson.In(
+            pyson.Eval('foo', 'test'), {'foo': 'bar'}))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
         eval = pyson.PYSONEncoder().encode(pyson.In('1', {2: 'bar'}))
         self.assertFalse(pyson.PYSONDecoder().decode(eval))
 
@@ -770,6 +792,18 @@ class PYSONTestCase(unittest.TestCase):
                 'm': 3,
                 })
 
+    def test_TimeDelta_enode(self):
+        "Test pyson.TimeDelta encode"
+        eval = pyson.PYSONEncoder().encode(pyson.TimeDelta(1, 2, 3))
+        self.assertEqual(
+            pyson.PYSONDecoder().decode(eval),
+            datetime.timedelta(1, 2, 3))
+
+        eval = pyson.PYSONEncoder().encode(datetime.timedelta(1, 2, 3))
+        self.assertEqual(
+            pyson.PYSONDecoder().decode(eval),
+            datetime.timedelta(1, 2, 3))
+
     def test_TimeDelta_types(self):
         "Test pyson.TimeDelta.types"
         self.assertEqual(
@@ -876,7 +910,3 @@ class PYSONTestCase(unittest.TestCase):
     def test_eval_false(self):
         "Test PYSON.eval JS false"
         self.assertEqual(eval('false', pyson.CONTEXT), False)
-
-
-def suite():
-    return unittest.TestLoader().loadTestsFromTestCase(PYSONTestCase)
