@@ -2,16 +2,76 @@
 # this repository contains the full copyright notices and license terms.
 import logging
 import time
-from collections import defaultdict
+from collections import defaultdict, deque
+from functools import wraps
 from threading import local
+from weakref import WeakValueDictionary
+
 from sql import Flavor
 
 from trytond.config import config
+from trytond.tools.immutabledict import ImmutableDict
 
+__all__ = ['Transaction',
+    'check_access', 'without_check_access',
+    'active_records', 'inactive_records']
+
+_retry = config.getint('database', 'retry')
+_cache_transaction = config.getint('cache', 'transaction')
 _cache_model = config.getint('cache', 'model')
+_cache_record = config.getint('cache', 'record')
 logger = logging.getLogger(__name__)
 
 
+class TransactionError(Exception):
+    def fix(self, extras):
+        pass
+
+
+class _TransactionLockError(TransactionError):
+    def __init__(self, table):
+        super().__init__()
+        self._table = table
+
+    def fix(self, extras):
+        super().fix(extras)
+        extras.setdefault('_lock_tables', []).append(self._table)
+
+
+def record_cache_size(transaction):
+    return transaction.context.get('_record_cache_size', _cache_record)
+
+
+def check_access(func=None, *, _access=True):
+    if func:
+        @wraps(func)
+        def wrapper(*args, **kwargs):
+            with check_access(_access=_access):
+                return func(*args, **kwargs)
+        return wrapper
+    else:
+        return Transaction().set_context(_check_access=_access)
+
+
+def without_check_access(func=None):
+    return check_access(func=func, _access=False)
+
+
+def active_records(func=None, *, _test=True):
+    if func:
+        @wraps(func)
+        def wrapper(*args, **kwargs):
+            with active_records(_test=_test):
+                return func(*args, **kwargs)
+        return wrapper
+    else:
+        return Transaction().set_context(active_test=_test)
+
+
+def inactive_records(func=None):
+    return active_records(func=func, _test=False)
+
+
 class _AttributeManager(object):
     '''
     Manage Attribute of transaction
@@ -45,24 +105,25 @@ class Transaction(object):
 
     cache_keys = {'language', 'fuzzy_translation', '_datetime'}
 
-    database = None
-    readonly = False
-    connection = None
-    close = None
-    user = None
-    context = None
-    create_records = None
-    delete_records = None
-    delete = None  # TODO check to merge with delete_records
-    trigger_records = None
-    timestamp = None
-    started_at = None
-
     def __new__(cls, new=False):
         transactions = cls._local.transactions
         if new or not transactions:
             instance = super(Transaction, cls).__new__(cls)
-            instance.cache = {}
+            instance.database = None
+            instance.readonly = False
+            instance.connection = None
+            instance.close = None
+            instance.user = None
+            instance.context = None
+            instance.create_records = None
+            instance.delete_records = None
+            instance.trigger_records = None
+            instance.log_records = None
+            instance.check_warnings = None
+            instance.timestamp = None
+            instance.started_at = None
+            instance.cache = WeakValueDictionary()
+            instance._cache_deque = deque(maxlen=_cache_transaction)
             instance._atexit = []
             transactions.append(instance)
         else:
@@ -82,51 +143,80 @@ class Transaction(object):
 
     def get_cache(self):
         from trytond.cache import LRUDict
+        from trytond.pool import Pool
         keys = tuple(((key, self.context[key])
                 for key in sorted(self.cache_keys)
                 if key in self.context))
-        return self.cache.setdefault((self.user, keys), LRUDict(_cache_model))
+        cache = self.cache.setdefault(
+            (self.user, keys), LRUDict(
+                _cache_model,
+                lambda name: LRUDict(
+                    record_cache_size(self),
+                    Pool().get(name)._record),
+                default_factory_with_key=True))
+        # Keep last used cache references to allow to pre-fill them
+        self._cache_deque.append(cache)
+        return cache
 
     def start(self, database_name, user, readonly=False, context=None,
-            close=False, autocommit=False):
+            close=False, autocommit=False, **extras):
         '''
         Start transaction
         '''
-        from trytond import backend
-        assert self.user is None
-        assert self.database is None
-        assert self.close is None
-        assert self.context is None
-        # Compute started_at before connect to ensure
-        # it is strictly before all transactions started after
-        # but it may be also before transactions started before
-        self.started_at = self.monotonic_time()
-        if not database_name:
-            database = backend.Database().connect()
-        else:
-            database = backend.Database(database_name).connect()
-        Flavor.set(backend.Database.flavor)
-        self.connection = database.get_connection(readonly=readonly,
-            autocommit=autocommit)
-        self.user = user
-        self.database = database
-        self.readonly = readonly
-        self.close = close
-        self.context = context or {}
-        self.create_records = {}
-        self.delete_records = {}
-        self.delete = {}
-        self.trigger_records = defaultdict(set)
-        self.timestamp = {}
-        self.counter = 0
-        self._datamanagers = []
-        if database_name:
-            from trytond.cache import Cache
-            try:
+        try:
+            from trytond import backend
+            assert self.user is None
+            assert self.database is None
+            assert self.close is None
+            assert self.context is None
+            # Compute started_at before connect to ensure
+            # it is strictly before all transactions started after
+            # but it may be also before transactions started before
+            self.started_at = self.monotonic_time()
+            if not database_name:
+                database = backend.Database().connect()
+            else:
+                database = backend.Database(database_name).connect()
+            Flavor.set(backend.Database.flavor)
+            self.user = user
+            self.database = database
+            self.readonly = readonly
+            self.close = close
+            self.context = ImmutableDict(context or {})
+            self.create_records = defaultdict(set)
+            self.delete_records = defaultdict(set)
+            self.trigger_records = defaultdict(set)
+            self.log_records = []
+            self.check_warnings = set()
+            self.timestamp = {}
+            self.counter = 0
+            self._datamanagers = []
+
+            count = 0
+            while True:
+                if count:
+                    time.sleep(0.002 * (_retry - count))
+                self.connection = database.get_connection(readonly=readonly,
+                    autocommit=autocommit)
+                try:
+                    lock_tables = extras.get('_lock_tables', [])
+                    for table in lock_tables:
+                        self.database.lock(self.connection, table)
+                    self._locked_tables = set(lock_tables)
+                except backend.DatabaseOperationalError:
+                    if count < _retry:
+                        self.connection.rollback()
+                        count += 1
+                        logger.debug("Retry: %i", count)
+                        continue
+                    raise
+                break
+            if database_name:
+                from trytond.cache import Cache
                 Cache.sync(self)
-            except BaseException:
-                self.stop(False)
-                raise
+        except BaseException:
+            self.stop(False)
+            raise
         return self
 
     def __enter__(self):
@@ -157,8 +247,8 @@ class Transaction(object):
                     self.context = None
                     self.create_records = None
                     self.delete_records = None
-                    self.delete = None
                     self.trigger_records = None
+                    self.log_records = None
                     self.timestamp = None
                     self._datamanagers = []
 
@@ -175,15 +265,16 @@ class Transaction(object):
         if context is None:
             context = {}
         manager = _AttributeManager(context=self.context)
-        self.context = self.context.copy()
-        self.context.update(context)
+        ctx = self.context.copy()
+        ctx.update(context)
         if kwargs:
-            self.context.update(kwargs)
+            ctx.update(kwargs)
+        self.context = ImmutableDict(ctx)
         return manager
 
     def reset_context(self):
         manager = _AttributeManager(context=self.context)
-        self.context = {}
+        self.context = ImmutableDict()
         return manager
 
     def set_user(self, user, set_context=False):
@@ -191,28 +282,47 @@ class Transaction(object):
             raise ValueError('set_context only allowed for root')
         manager = _AttributeManager(user=self.user,
                 context=self.context)
-        self.context = self.context.copy()
+        ctx = self.context.copy()
         if set_context:
             if user != self.user:
-                self.context['user'] = self.user
+                ctx['user'] = self.user
         else:
-            self.context.pop('user', None)
+            ctx.pop('user', None)
+        self.context = ImmutableDict(ctx)
         self.user = user
         return manager
 
+    def lock_table(self, table):
+        if table not in self._locked_tables:
+            raise _TransactionLockError(table)
+
     def set_current_transaction(self, transaction):
         self._local.transactions.append(transaction)
         return transaction
 
-    def new_transaction(self, autocommit=False, readonly=False):
+    def new_transaction(self, autocommit=False, readonly=False, **extras):
         transaction = Transaction(new=True)
         return transaction.start(self.database.name, self.user,
             context=self.context, close=self.close, readonly=readonly,
-            autocommit=autocommit)
+            autocommit=autocommit, **extras)
+
+    def _store_log_records(self):
+        from trytond.pool import Pool
+        if self.log_records:
+            pool = Pool()
+            Log = pool.get('ir.model.log')
+            with without_check_access():
+                Log.save(self.log_records)
+        self._clear_log_records()
+
+    def _clear_log_records(self):
+        if self.log_records:
+            self.log_records.clear()
 
     def commit(self):
         from trytond.cache import Cache
         try:
+            self._store_log_records()
             if self._datamanagers:
                 for datamanager in self._datamanagers:
                     datamanager.tpc_begin(self)
@@ -244,6 +354,7 @@ class Transaction(object):
         for datamanager in self._datamanagers:
             datamanager.tpc_abort(self)
         Cache.rollback(self)
+        self._clear_log_records()
         self.connection.rollback()
 
     def join(self, datamanager):
@@ -266,3 +377,11 @@ class Transaction(object):
         if self.context:
             return self.context.get('language') or get_language()
         return get_language()
+
+    @property
+    def check_access(self):
+        return self.context.get('_check_access', False)
+
+    @property
+    def active_records(self):
+        return self.context.get('active_test', True)
